"""Package containing io convenience functionality."""

from __future__ import annotations

from pyglotaran_extras.io.load_data import load_data
from pyglotaran_extras.io.setup_case_study import setup_case_study

__all__ = ["setup_case_study", "load_data"]
